unit PolynomUtil;

{$mode objfpc}{$H+}


interface

uses
  Classes, SysUtils, StrTransformErr, typinfo, Dialogs;

type
 generic Polynom<Precision> = class
  type
   Complex = object
    public
     Re : Precision;
     Im : Precision;
   end;

   IFunction = class
     function Calc(x : Precision) : Precision; virtual; abstract;
     function Derivate : IFunction; virtual; abstract;
   end;

   TPolynom = class(IFunction)
    private
     _co         : Array of Precision;
     _ze         : Array of Precision;
     _grad       : Cardinal;
     procedure calcCoefficients;
     procedure calcZeros;
     function ReadCo(i : Cardinal) : Precision; inline;
     procedure WriteCo(i : Cardinal; v : Precision);
    protected
     destructor Destroy; override;
    public
     constructor Create;
     function ToString : String;
     property co[i : Cardinal] : Precision read ReadCo write WriteCo; default;
     property Grad : Cardinal read _grad;
     function Calc(x : Precision) : Precision; override;
     procedure Add(P : TPolynom);
     function Derivate : TPolynom; override;
   end;

   TPolySpline = class(IFunction)
    type TPolySplinePart = record
     Fkt : TPolynom;
     Min : Precision;
     Max : Precision;
    end;
    private
     _Splines   : Array of TPolySplinePart;
     _LastIndex : Cardinal;
    protected
     destructor Destroy; override;
    public
     constructor Create;
     function Calc(x : Precision) : Precision; override;
     function Derivate : TPolySpline; override;
   end;

   TInterpolatePoint = record
    x : Precision;
    y : Precision;
    derivation : Cardinal;
   end;

   TMatrix = object
    private
     Mem  : Array of Precision;
     dim1 : Cardinal;
     dim2 : Cardinal;
     function _get(i, j : Cardinal) : Precision; inline;
     procedure _set(i, j : Cardinal; v : Precision); inline;
    public
     constructor Init(i, j : Cardinal; Identity : Boolean);
     property Spalten : Cardinal read dim2;
     property Zeilen : Cardinal read dim1;
     property Objects[i, j : Cardinal] : Precision read _get write _set; default;
   end;

   TInterpolateInfo = Array of TInterpolatePoint;

  private
   class function FloatToStr(f : Precision) : String; inline;
   class procedure SortInterpolatePoints(var A : Array of TInterpolatePoint; _lo, _hi : integer);

  public
   class function NewtonInterpolation(const IP : Array of TInterpolatePoint) : TPolynom;
   class function AdjustedSplinesOnFirstDerivation(var IP : Array of TInterpolatePoint; kompensate : Precision) : TPolySpline;
   class function KubicSplines(var IP : Array of TInterpolatePoint) : TPolySpline;
   class function IsNaN(v : Precision) : Boolean; inline;
   class function NaN : Precision; inline;
 end;



implementation

constructor Polynom.TMatrix.Init(i, j : Cardinal; Identity : Boolean);
begin
 SetLength(Mem, i * j);
 dim2 := j;
 dim1 := i;
 while (i > 0) do begin
  i := i - 1;
  while (j > 0) do begin
   j := j - 1;
   if (Identity) and (i = j) then Mem[i * dim2 + j] := 1
   else Mem[i * dim2 + j] := 0;
  end;
 end;
end;

function Polynom.TMatrix._get(i, j : Cardinal) : Precision; inline;
begin
 Result := Mem[i * dim2 + j];
end;

procedure Polynom.TMatrix._set(i, j : Cardinal; v : Precision) inline;
begin
 Mem[i * dim2 + j] := v;
end;




constructor Polynom.TPolynom.Create;
begin
 inherited;
 SetLength(_co, 1);
 _co[0] := 0;
 SetLength(_ze, 0);
 _grad := 0;
end;

procedure Polynom.TPolynom.WriteCo(i : Cardinal; v : Precision);
begin
 if (i > _grad) then begin
  SetLength(_co, i+1);
  repeat begin _grad := _grad + 1; _co[_grad] := 0; end until (_grad >= i);
 end;
 _co[i] := v;
 if (i = _grad) and (v = 0) then while (_grad > 0) and (_co[i] = 0) do begin
  SetLength(_co, _grad+1);
  _grad := _grad - 1;
 end;
end;

function Polynom.TPolynom.ReadCo(i : Cardinal) : Precision; inline;
begin
 if (i > _grad) then Result := 0
 else Result := _co[i];
end;

procedure Polynom.TPolynom.calcCoefficients;
begin

end;

procedure Polynom.TPolynom.calcZeros;
begin

end;

function Polynom.TPolynom.ToString : String;
var i : integer;
begin
 Result := '';
 //if (not _coValid) then calcCoefficients;
 i := Length(_co)-1;
 for i := i downto 0 do begin
  if (_co[i] < 0) then Result := Result + ' - ' + FloatToStr(-_co[i]) + ' X^' + IntToStr(i)
  else Result := Result + ' + ' + FloatToStr(_co[i]) + ' X^' + IntToStr(i);
 end;
end;

function Polynom.TPolynom.Calc(x : Precision) : Precision;
var
 Ex : Precision;
 i  : integer;
begin
 Ex := 1;
 Result := 0;
 for i := 0 to _grad do begin
  Result := Result + _co[i] * Ex;
  Ex := Ex * x;
 end;
end;

procedure Polynom.TPolynom.Add(P : TPolynom);
var i : integer;
begin
 if (P._grad > _grad) then begin
  SetLength(_co, P._grad+1);
  repeat begin _grad := _grad + 1; _co[_grad] := 0; end until (_grad >= P._grad);
 end;
 for i := P._grad downto 0 do _co[i] := _co[i] + P._co[i];
end;

function Polynom.TPolynom.Derivate : TPolynom;
var i : integer;
begin
 Result := TPolynom.Create;
 for i := _grad downto 1 do Result[i-1] := _co[i]*i;
end;

destructor Polynom.TPolynom.Destroy;
begin
 SetLength(_co, 0);
 SetLength(_ze, 0);
 inherited;
end;



constructor Polynom.TPolySpline.Create;
begin
 inherited;
 _LastIndex := 0;
end;

function Polynom.TPolySpline.Calc(x : Precision) : Precision;
var
  i    : Cardinal;
  b    : Boolean;
  l, u : Cardinal;
begin
 Result := NaN;
 if (Length(_Splines) = 0) then exit;
 i := _LastIndex;
 l := 0;
 u := Length(_Splines)-1;
 if (i >= Length(_Splines)) then i := (Length(_Splines)-1) div 2;
 b := false;
 repeat begin
  if (_Splines[i].Min > x) then begin
   if (i = l) then exit;
   u := i - 1;
   i := (u + l) div 2;
  end else if (_Splines[i].Max <= x) then begin
   if (i = u) then exit;
   l := i + 1;
   i := (u + l) div 2;
  end else b := true;
 end until (b);
 _LastIndex := i;
 Result := _Splines[i].Fkt.Calc(x);
end;

function Polynom.TPolySpline.Derivate : TPolySpline;
var i : integer;
begin
 Result := TPolySpline.Create;
 SetLength(Result._Splines, Length(_Splines));
 for i := 0 to Length(_Splines)-1 do begin
  Result._Splines[i].Min := _Splines[i].Min;
  Result._Splines[i].Max := _Splines[i].Max;
  Result._Splines[i].Fkt := _Splines[i].Fkt.Derivate;
 end;
end;

destructor Polynom.TPolySpline.Destroy;
var i : integer;
begin
 for i := Length(_Splines)-1 downto 0 do _Splines[i].Fkt.Free;
 SetLength(_Splines, 0);
 inherited;
end;




class function Polynom.FloatToStr(f : Precision) : String; inline;
begin
 Result := '?';
 case (PTypeInfo(TypeInfo(Precision))^.Kind) of
  tkFloat : begin
    case (GetTypeData(PTypeInfo(TypeInfo(Precision)))^.FloatType) of
     ftSingle   : Result := SysUtils.FloatToStr(Single(f));
     ftDouble   : Result := SysUtils.FloatToStr(Double(f));
     ftExtended : Result := SysUtils.FloatToStr(Extended(f));
    end;
  end;
 end;
end;

class function Polynom.NewtonInterpolation(const IP : Array of TInterpolatePoint) : TPolynom;
var
  i, j, l : integer;
begin
 Result := TPolynom.Create;
 l := Length(IP)-1;
 for i := 0 to l do begin
  //if (IP[l].derivation <> 0) then Exception.Create('Newton-Interpolation must not contain function values of derivations');

 end;
 Result[0] := 1;
 Result[2] := -1;
 Result[3] := 0.1;
end;

class procedure Polynom.SortInterpolatePoints(var A : Array of TInterpolatePoint; _lo, _hi : integer);
var
 lo,hi : integer;
 m,s : TInterpolatePoint;
begin
 if (Length(A) = 0) then exit;
 m := A[(_lo + _hi) div 2];
 lo := _lo;
 hi := _hi;
 repeat
  while (A[lo].x < m.x) do lo := lo+1;
  while (A[hi].x > m.x) do hi := hi-1;
  if (lo <= hi) then begin
   s := A[lo];
   A[lo] := A[hi];
   A[hi] := s;
   lo := lo+1;
   hi := hi-1;
  end;
 until (lo > hi);
 if (hi > _lo) then SortInterpolatePoints(A,_lo,hi);
 if (lo < _hi) then SortInterpolatePoints(A,lo,_hi);
end;

class function Polynom.AdjustedSplinesOnFirstDerivation(var IP : Array of TInterpolatePoint; kompensate : Precision) : TPolySpline;
var
 n, _l, r, s : integer;
 k, ku, ko   : Precision;

 i, j, l, m : integer;
 mx, my     : Precision;

begin
 Result := TPolySpline.Create;
 SortInterpolatePoints(IP, 0, Length(IP)-1);
 _l := -1;
 s := 0;
 for n := 0 to Length(IP)-1 do if (IP[n].derivation = 0) then begin
  ku := 0.5 - kompensate;
  ko := 0.5 + kompensate;

  for r := 1 to trunc(kompensate * 5) + 1 do begin
   k := (ku + ko) / 2;
   SetLength(Result._Splines, s);
   j := -1;

   l := _l;
   for i := _l+1 to Length(IP)-1 do if (IP[i].derivation = 1) then begin
    if (l = -1) then begin l := i; continue; end;
    j := Length(Result._Splines);
    SetLength(Result._Splines, j+1);
    Result._Splines[j].Fkt := TPolynom.Create;
    mx := (IP[i].x + IP[l].x)/2;
    if (IP[i].y - IP[l].y >= 0) then my := IP[l].y + (IP[i].y - IP[l].y) * k
    else my := IP[l].y + (IP[i].y - IP[l].y) * (1-k);
    Result._Splines[j].Fkt[3] := ((IP[i].x*(IP[l].y - my) + IP[i].y*(mx - IP[l].x) + IP[l].x*my - IP[l].y*mx)/((IP[i].x*IP[i].x - IP[i].x*(IP[l].x + mx) + IP[l].x*mx)*(mx - IP[l].x)))/3;
    Result._Splines[j].Fkt[2] := ((IP[i].x*IP[i].x*(IP[l].y - my) + IP[i].y*(IP[l].x + mx)*(mx - IP[l].x) + IP[l].x*IP[l].x*my - IP[l].y*mx*mx)/((IP[i].x*IP[i].x - IP[i].x*(IP[l].x + mx) + IP[l].x*mx)*(IP[l].x - mx)))/2;
    Result._Splines[j].Fkt[1] := (IP[i].x*IP[i].x*(IP[l].x*my - IP[l].y*mx) + IP[i].x*(IP[l].y*mx*mx - IP[l].x*IP[l].x*my) + IP[i].y*IP[l].x*mx*(IP[l].x - mx))/((IP[i].x*IP[i].x - IP[i].x*(IP[l].x + mx) + IP[l].x*mx)*(IP[l].x - mx));
    Result._Splines[j].Fkt[0] := 0;
    if (j = 0) then begin
     for m := 0 to Length(IP)-1 do if (IP[m].derivation = 0) then begin
      Result._Splines[0].Fkt[0] := - Result._Splines[0].Fkt.Calc(IP[m].x) + IP[m].y;
      break;
     end;
    end else Result._Splines[j].Fkt[0] := Result._Splines[j-1].Fkt.Calc(IP[l].x) - Result._Splines[j].Fkt.Calc(IP[l].x);
    Result._Splines[j].Min := IP[l].x;
    Result._Splines[j].Max := IP[i].x;
    l := i;
    if (IP[i].x >= IP[n].x) then break;
   end;

   if (j <= 0) then break;
   if (Result._Splines[j].Fkt.Calc(IP[n].x) > IP[n].y) then ko := k else ku := k;
  end;

  s := Length(Result._Splines);
  _l := l;
 end;
end;

class function Polynom.KubicSplines(var IP : Array of TInterpolatePoint) : TPolySpline;
begin
 // ToDo
end;

class function Polynom.IsNaN(v : Precision) : Boolean;
begin
 case (PTypeInfo(TypeInfo(Precision))^.Kind) of
   tkFloat : begin
     case (GetTypeData(PTypeInfo(TypeInfo(Precision)))^.FloatType) of
      ftSingle   : Result := v = $FFFFFFFF;
      ftDouble   : Result := v = $FFFFFFFFFFFFFFFF;
      ftExtended : Result := (PQWord(@v)^ = $FFFFFFFFFFFFFFFF) and (PWord(@v + 8)^ = $FFFF);
     end;
   end;
  end;
end;

class function Polynom.NaN : Precision;
begin
 case (PTypeInfo(TypeInfo(Precision))^.Kind) of
   tkFloat : begin
     case (GetTypeData(PTypeInfo(TypeInfo(Precision)))^.FloatType) of
      ftSingle   : Result := $FFFFFFFF;
      ftDouble   : Result := $FFFFFFFFFFFFFFFF;
      ftExtended : begin PQWord(@Result)^ := $FFFFFFFFFFFFFFFF; PWord(@Result + 8)^ := $FFFF; end;
     end;
   end;
  end;
end;

end.

